/*! \file usb_small_lib_hid_keyboard_firmware_v1_2_c.c
********************************************************************************
 main source for using the small usb library using hid class
 HID keyboard 
 Firmware zum Testen der HID Klasse der kleinen USB-Bibliothek
 Tastatur-Emulation
 
 www.weigu.lu/b/usb

 (inspired by the lib from S. Salewski (http://http://www.ssalewski.de)
 
 \author weigu www.weigu.lu

        Controller: AT90USB162, AT90USB1287, ATmega32u4, ATmega32u2
                   (siehe Zuweisungen makefile)
        Quarz:     8 oder 16 MHz  (siehe Zuweisungen makefile)

        In der Makefile muss der Chip und die Quarzfrequenz für das
        entsprechende Board festgelegt werden!

        Teensy 2.0 (ATmega32u4, 16 MHz, www.pjrc.com/teensy):
        
        Eingaenge:   PB0-7 (8 Tasten)
        Ausgaenge:   PD6 (LED on board)

        AT90USBKEY (AT90USB1287, 8 MHz, www.atmel.com/tools/AT90USBKEY.aspx):
        
        Eingaenge:   PB0-7 (8 Tasten) ev. anpassen
        Ausgaenge:   PD4-7 (LEDs on board, PD6 blinkt)

        Teensy 1.0 (AT90USB162, 16 MHz, www.pjrc.com/teensy):

        Eingaenge:   PB0-7 (8 Tasten)
        Ausgaenge:   PD6 (LED on board)
        
        Informationen zur Funktionsweise:

        Hauptprogramm zur Nutzung der HID-Bibliothek
        Ein Timerinterrupt eine LED an PD6 mit 1 Hz blinken.
        Die gesamte Arbeit zum USB-Bus wird mittels Interrupt-Service-Routinen
        abgewickelt. Alle USB-Routinen und Unterprogramme befinden sich in der
        USB-Bibliothek.
        
        Fuer beide Endpunkte steht ein 64-Byte großer Buffer zur Verfuegung. 

        Beim Endpunkt 1 (IN) fragt der PC Daten an.
        Durch Setzen eines Flags (EP1_FLAG) wird dem Endpunkt mitgeteilt, wenn
        neue Daten vorliegen. Die Anzahl der vorhandenen Bytes steht in der
        Variablen EP1_CNT. Die Daten selbst werden im Buffer (EP1_BUF) abgelegt.
        
        Ueber den Endpunkt 2 (OUT) sendet der PC Daten.
        Die Hauptschleife erkennt an einem Flag (EP2_FLAG) wenn Daten vorliegen.
        Die Daten liegen im Buffer (EP2_BUF). Die Variable EP2_CNT gibt an wie 
        viele Bytes gesendet wurden.
        
        Es wird nach 5 Sekunden Wartezeit einmalig der Text "Hallo" gesendet.
        Danach werden die 8 Bit (Tasten) von Port B eingelesen (nullaktiv).
        Wurde eine Taste betätigt, so wird ein Zeichen gesendet.
        
        Das Freischalten der Tastatur-Emulation erfolgt in der Headerdatei
        lib/usb_srs_hid_v1_2.h
      
        Weitere Dokumentation auf www.weigu.lu/b/usb


******************************************************************************/


/* Copyright (c) 2013   Guy WEILER         weigu[at]weigu[dot]lu

   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>.

   Licence: GPL-3        */

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include "lib/usb_srs_hid_v1_2.h"

/*! \brief  Timer1 Compare Match A Interrupt
    Toggelt LED an PD6 */
ISR (TIMER1_COMPA_vect)
{
  SBI (PIND,6);       //Toggelt bei neuen Chips das Pin!  
}

/*! \brief Unterprogramm zum Senden eines Zeichens 
    \PARAM mod:  byte for modifier key, Bit 7-0:
                 R GUI, R ALT, R SHIFT, R CTRL,L GUI, L ALT, L SHIFT, L CTRL
           code: scan code from key  */

void sendkey(uint8_t mod, uint8_t code)
{
  Ep1_cnt = 8;        // 8 bytes for Windows
  cli();
  Ep1_buf[0] = mod;   // modifier byte
  Ep1_buf[1] = 0x00;  // reserved byte = 0
  Ep1_buf[2] = code;  // scan code (up to 6 keys)
  Ep1_buf[3] = 0x00;  // reserved byte = 0
  Ep1_buf[4] = 0x00;  // reserved byte = 0
  Ep1_buf[5] = 0x00;  // reserved byte = 0
  Ep1_buf[6] = 0x00;  // reserved byte = 0
  Ep1_buf[7] = 0x00;  // reserved byte = 0
  sei();
  Ep1_flag = 1;       // set EP1 flag
  _delay_ms(5);       // higher than bInterval in Endpoint Descriptor!
  Ep1_cnt = 8;        // 8 bytes
  cli();
  Ep1_buf[0] = 0x00;  // modifier byte
  Ep1_buf[1] = 0x00;  // reserved byte = 0
  Ep1_buf[2] = 0x00;  // stop pressing key
  Ep1_buf[3] = 0x00;  // reserved byte = 0
  Ep1_buf[4] = 0x00;  // reserved byte = 0
  Ep1_buf[5] = 0x00;  // reserved byte = 0
  Ep1_buf[6] = 0x00;  // reserved byte = 0
  Ep1_buf[7] = 0x00;  // reserved byte = 0
  sei();
  Ep1_flag = 1;       // set EP1 flag
  _delay_ms(5); 
}        

/*! \brief Hauptprogramm
 */
int main(void)
{
  // Wirkung des Fusebit CKDIV ueberschreiben
  CLKPR = (1<<CLKPCE);  
  CLKPR = 0;
  // Timer1 initialisieren, Interrupt ein mal pro Sekunde
  // Vorteiler 1024, Zaehlschritte 7811-1 ergeben bei 16MHz 1 Sekunde
  TCCR1A = 0;
  TCCR1B = 0x0D;                // WGM12 = 1 Teiler = 1024
#if F_CPU == 8000000
  OCR1AH = high(3905);          // Achtung High zuerst!
  OCR1AL = low(3905);
#else
  OCR1AH = high(7811);          // Achtung High zuerst!
  OCR1AL = low(7811);
#endif
  TIMSK1 |= (1 << OCIE1A);      // OCI1A Interrupt erlauben
  
  // Eingänge festlegen (! Anpassen je nach Chip)
  DDRB = 0;
  PORTB = 0xFF;         // Pull-Ups
  
  //Ausgaenge initialisieren
  DDRD |= (1<<DDD6); // PD6 = Ausgang

  // USB starten (erlaubt auch Interrupts (sei))
  usb_init_device();    // Unterprogramm der USB-Bibliothek
  
  //Warte 5 Sekunden
   _delay_ms(5000);
   
  // Schreibe "Hallo" 
  sendkey(0x02,0x0B); //H (Shift+h)
  sendkey(0x00,0x04); //a
  sendkey(0x00,0x0F); //l
  sendkey(0x00,0x0F); //l
  sendkey(0x00,0x12); //o
  sendkey(0x00,0x2C); //Space
  
  while (1)
  {        
    // 8 Tasten abfragen      
    if ((PINB & (1<<PB0)) == 0) sendkey(0x00,0x14); //q
    if ((PINB & (1<<PB1)) == 0) sendkey(0x00,0x1A); //w
    if ((PINB & (1<<PB2)) == 0) sendkey(0x00,0x08); //e
    if ((PINB & (1<<PB3)) == 0) sendkey(0x00,0x15); //r
    if ((PINB & (1<<PB4)) == 0) sendkey(0x00,0x17); //t
    if ((PINB & (1<<PB5)) == 0) sendkey(0x00,0x1C); //z
    if ((PINB & (1<<PB6)) == 0) sendkey(0x00,0x18); //u
    if ((PINB & (1<<PB7)) == 0) sendkey(0x00,0x0C); //i
    _delay_ms(100); 
  }
}
